unit CelMaterial;

interface

uses
  System.SysUtils, System.Classes, System.Types, System.UITypes, FMX.Types,
  FMX.Types3D, FMX.Materials, FMX.MaterialSources;

type
  TCelMaterial = class(TCustomMaterial)
  private
    FThreshold: Single;
    FDarkColor: TAlphaColor;
    FLightColor: TAlphaColor;
    [Weak] FTexture: TTexture;
    procedure SetThreshold(Value: Single);
    procedure SetDarkColor(const Value: TAlphaColor);
    procedure SetLightColor(const Value: TAlphaColor);
  protected
    procedure DoApply(const Context: TContext3D); override;
    procedure DoInitialize; override;
  public
    constructor Create; override;

    property DarkColor: TAlphaColor read FDarkColor write SetDarkColor;
    property LightColor: TAlphaColor read FLightColor write SetLightColor;
    property Threshold: Single read FThreshold write SetThreshold;
    property Texture: TTexture read FTexture write FTexture;
  end;

  TCelMaterialSource = class(TMaterialSource)
  private
    FTexture: TBitmap;
    function GetDarkColor: TAlphaColor;
    function GetLightColor: TAlphaColor;
    function GetThreshold: Single;
    procedure SetDarkColor(const Value: TAlphaColor);
    procedure SetLightColor(const Value: TAlphaColor);
    procedure SetThreshold(const Value: Single);
    procedure SetTexture(const Value: TBitmap);
    procedure DoTextureChanged(Sender: TObject);
  protected
    function CreateMaterial: TMaterial; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property DarkColor: TAlphaColor read GetDarkColor write SetDarkColor;
    property LightColor: TAlphaColor read GetLightColor write SetLightColor;
    property Threshold: Single read GetThreshold write SetThreshold;
    property Texture: TBitmap read FTexture write SetTexture;
  end;

implementation

{ TCelMaterial }

constructor TCelMaterial.Create;
begin
  inherited;
  FDarkColor := $FF303030;
  FLightColor := $FFF0F0F0;
  FThreshold := 0.3;
end;

procedure TCelMaterial.DoApply(const Context: TContext3D);
var
  i: Integer;
  hasLight: Boolean;
begin
  inherited;
  Context.SetShaderVariable('Color1', FDarkColor);
  Context.SetShaderVariable('Color2', FLightColor);
  if (FTexture = nil) or FTexture.IsEmpty then
    Context.SetShaderVariable('Option', [Vector3D(FThreshold, 0, 0, 0)])
  else begin
    Context.SetShaderVariable('Option', [Vector3D(FThreshold, 1, 0, 0)]);
    Context.SetShaderVariable('Tex', FTexture);
  end;

  hasLight := False;
  for i := 0 to Context.Lights.Count-1 do
  begin
    if Context.Lights[i].Enabled then
    begin
      Context.SetShaderVariable('LightDirection', [Context.Lights[i].Direction]);
      Context.SetShaderVariable('LightColor', Context.Lights[i].Color);
      hasLight := True;
      Break;
    end;
  end;
  if not hasLight then
  begin
    Context.SetShaderVariable('LightDirection', [Vector3D(0, 0, 0, 0)]);
    Context.SetShaderVariable('LightColor', [Vector3D(0, 0, 0, 1)]);
  end;
end;

procedure TCelMaterial.DoInitialize;

  function BytesFromFile(const FileName: string): TBytes;
  var
    ms: TMemoryStream;
  begin
    ms := TMemoryStream.Create;
    try
      ms.LoadFromFile(FileName);
      SetLength(Result, ms.Size);
      ms.ReadBuffer(Pointer(Result)^, ms.Size);
    finally
      ms.Free;
    end;
  end;

begin
  inherited;
  FVertexShader := TShaderManager.RegisterShaderFromData('cel.vs', TContextShaderKind.skVertexShader, '', [
    TContextShaderSource.Create(TContextShaderArch.saDX9, BytesFromFile('celDX9.cvs'), [
      TContextShaderVariable.Create('MVPMatrix', TContextShaderVariableKind.vkMatrix, 0, 4),
      TContextShaderVariable.Create('LightDirection', TContextShaderVariableKind.vkVector, 4, 1)
    ])
  ]);
  FPixelShader := TShaderManager.RegisterShaderFromData('cel.ps', TContextShaderKind.skPixelShader, '', [
    TContextShaderSource.Create(TContextShaderArch.saDX9, BytesFromFile('celDX9.cps'), [
      TContextShaderVariable.Create('Color1', TContextShaderVariableKind.vkVector, 0, 1),
      TContextShaderVariable.Create('Color2', TContextShaderVariableKind.vkVector, 1, 1),
      TContextShaderVariable.Create('Option', TContextShaderVariableKind.vkVector, 2, 1),
      TContextShaderVariable.Create('LightColor', TContextShaderVariableKind.vkVector, 3, 1),
      TContextShaderVariable.Create('Tex', TContextShaderVariableKind.vkTexture, 0, 0)
    ])
  ]);
end;

procedure TCelMaterial.SetDarkColor(const Value: TAlphaColor);
begin
  if FDarkColor <> Value then
  begin
    FDarkColor := Value;
    DoChange;
  end;
end;

procedure TCelMaterial.SetLightColor(const Value: TAlphaColor);
begin
  if FLightColor <> Value then
  begin
    FLightColor := Value;
    DoChange;
  end;
end;

procedure TCelMaterial.SetThreshold(Value: Single);
begin
  if Value < 0.0 then
    Value := 0.0
  else if Value > 1.0 then
    Value := 1.0;

  if FThreshold <> Value then
  begin
    FThreshold := Value;
    DoChange;
  end;
end;

{ TCelMaterialSource }

constructor TCelMaterialSource.Create(AOwner: TComponent);
begin
  inherited;
  FTexture := TBitmap.Create(0, 0);
  FTexture.OnChange := DoTextureChanged;
end;

destructor TCelMaterialSource.Destroy;
begin
  FreeAndNil(FTexture);
  inherited;
end;

function TCelMaterialSource.CreateMaterial: TMaterial;
begin
  Result := TCelMaterial.Create;
end;

procedure TCelMaterialSource.DoTextureChanged(Sender: TObject);
begin
  if not FTexture.IsEmpty then
    TCelMaterial(Material).Texture := FTexture.Texture;
end;

function TCelMaterialSource.GetDarkColor: TAlphaColor;
begin
  Result := TCelMaterial(Material).DarkColor;
end;

function TCelMaterialSource.GetLightColor: TAlphaColor;
begin
  Result := TCelMaterial(Material).LightColor;
end;

function TCelMaterialSource.GetThreshold: Single;
begin
  Result := TCelMaterial(Material).Threshold;
end;

procedure TCelMaterialSource.SetDarkColor(const Value: TAlphaColor);
begin
  TCelMaterial(Material).DarkColor := Value;
end;

procedure TCelMaterialSource.SetLightColor(const Value: TAlphaColor);
begin
  TCelMaterial(Material).LightColor := Value;
end;

procedure TCelMaterialSource.SetTexture(const Value: TBitmap);
begin
  FTexture.Assign(Value);
end;

procedure TCelMaterialSource.SetThreshold(const Value: Single);
begin
  TCelMaterial(Material).Threshold := Value;
end;

initialization
  RegisterFmxClasses([TCelMaterialSource]);
end.
